from numpy import true_divide
from ppinot4py.model import AppliesTo, RuntimeState, TimeInstantCondition
import pandas as pd


def condition_computer(dataframe, id_case, condition, activity_column, transition_column):

    if condition is None:
        filteredSeries = dataframe
        
    if(type(condition) == str):
        dataframeValue = dataframe.query(condition)
        filteredArray = dataframe.index.isin(dataframeValue.index)
        filteredSeries = pd.Series(filteredArray)

    if(type(condition) == TimeInstantCondition):
        filteredSeries = _time_instant_condition_resolve(dataframe, id_case, condition, activity_column, transition_column)
        
    #In revision    
    if(type(condition) == pd.Series):
        filteredSeries = condition
  
    return filteredSeries


def _time_instant_condition_resolve(dataframe, id_case, condition: TimeInstantCondition, activity_column, transition_column):
    TimeInstantCondition(RuntimeState.START, AppliesTo.PROCESS)

    if condition.applies_to == AppliesTo.DATA:
        return _time_instant_condition_data_resolve(dataframe, id_case, str(condition.changes_to_state))
    elif condition.applies_to == AppliesTo.PROCESS:
        return _time_instant_condition_process_resolve(dataframe, id_case, condition.changes_to_state)
    elif condition.applies_to == AppliesTo.ACTIVITY:
        condition_checked = check_State(condition.changes_to_state)
        if(check_column_existence(dataframe, activity_column, transition_column)):
            data_condition = f'`{activity_column}` == {condition.activity_name} and `{transition_column}` == {condition_checked}'
            return _time_instant_condition_data_resolve(dataframe, id_case, data_condition)
        elif(check_end_condition(condition.changes_to_state)):
            data_condition = f'`{activity_column}` == {condition.activity_name}'
            return _time_instant_condition_data_resolve(dataframe, id_case, data_condition)
        else: 
            raise ValueError("The activity or transition column don't exists in log")
    else:
        raise ValueError("invalid applies to condition " + str(condition.applies_to))

def _time_instant_condition_process_resolve(dataframe, id_case, changes_to_state):
    default_false = pd.Series(data=False, index=dataframe.index)

    if changes_to_state == RuntimeState.START:
        shift = +1
    else:
        shift = -1

    condition = default_false.groupby(dataframe[id_case]).shift(shift).fillna(True)
    return condition

def _time_instant_condition_data_resolve(dataframe, id_case, var):
    condition = dataframe.query(var)

    condition_in_series = pd.Series(dataframe.index.isin(condition.index))

    condition_in_series_prev = condition_in_series.groupby(dataframe[id_case]).shift(+1).fillna(False)

    final_evaluation = ((condition_in_series) & (~condition_in_series_prev))

    # id_next = dataframe[id_case].shift(-1)
    # id_pre = dataframe[id_case].shift(+1)
    # condition = dataframe.query(var)

    # is_last = (dataframe[id_case] != dataframe[id_case].shift(-1)).reset_index(drop=True)
    # prev_is_last = is_last.shift(+1).fillna(True)

    # dataframeValue = dataframe.index.isin(condition.index)

    # conditionInSeries = pd.Series(dataframeValue)
    # conditionInSeries_prev = conditionInSeries.shift(+1).fillna(False)

    # partial_evaluation_next = conditionInSeries.shift(+1).fillna(False)

    # final_evaluation = ((conditionInSeries) & ((~conditionInSeries_prev) | prev_is_last) )

    # final_evaluation = ((conditionInSeries == True) &
    #                         (partial_evaluation_next == False) & 
    #                             ((dataframe[id_case] == id_pre) | (dataframe[id_case] == id_next)))
    # I don't think this == id_next is ok. It doesn't meet the condition if the process instance
    # finishes

    return final_evaluation    


def check_column_existence(dataframe, activity_column, transition_column):
    return (activity_column and transition_column in dataframe.columns)


def check_end_condition(condition):
    return condition == RuntimeState.END


def check_State(condition):
    if(condition == RuntimeState.START):
        return "'Assigned'"
    elif(condition == RuntimeState.END):
        return "'Completed'"
    else:
        return condition



from .conditions import (
    TimeInstantCondition, 
    AppliesTo,
    DataCondition, 
    SeriesCondition
)

from .measures import (
    CountMeasure,
    TimeMeasure,
    DataMeasure,
    AggregatedMeasure, 
    DerivedMeasure,
    BusinessDuration,
    RollingWindow,
    TimeUnit
)

from .states import (
    DataObjectState,
    RuntimeState
)

__all__ = [
    'TimeInstantCondition',
    'AppliesTo',
    'DataCondition',
    'SeriesCondition',
    'CountMeasure',
    'TimeMeasure',
    'DataMeasure',
    'AggregatedMeasure',
    'DerivedMeasure',
    'TimeUnit',
    'BusinessDuration',
    'RollingWindow',
    'DataObjectState',
    'RuntimeState'
]


from .conditions_computer import condition_computer
from .metrics_computer import measure_computer


from .metrics_computer import(
    LogConfiguration
)
__all__ = [
    'condition_computer',
    'measure_computer',
    'LogConfiguration'
]


from .conditions import TimeInstantCondition
import datetime
import enum

def _time_instance_auto_wrap(condition):
    if condition is None:
        return None

    if isinstance(condition, str):
        condition = TimeInstantCondition(condition)
    
    return condition

def _name_of_aggregation(agg):
    if agg == "AVG":
        return "average"
    if agg == "MIN":
        return "minimum"
    if agg == "MAX":
        return "maximum"
    if agg == "SUM":
        return "sum"

class TimeUnit(enum.Enum):
    WEEK = "W"
    DAY = "D"
    HOUR = "h"
    MINUTE = "m"
    SECOND = "s"

class _MeasureDefinition():
    def __init__(self):
        super().__init__()
        self.id = None

class CountMeasure(_MeasureDefinition):
    """Measure to calc how many times a condition happened.
    
    Args:
    
            - when: Condition that you want to count
    """

    def __init__(self, when):
        super().__init__()
        self.when = _time_instance_auto_wrap(when)
    
    def __repr__(self):
        return f"CountMeasure ( {self.when} )"

    def __str__(self):
        return f"the number of times {self.when}"

class DataMeasure(_MeasureDefinition):
    """Measure that returns a data from specific filter.
    
    Args:
    
            - data_content_selection: Column you want to select
            - precondition: Condition (Time Instance Condition by default) to apply to data
            - first: True if you want to select first data, False if you want to select last
    """
    
    def __init__(self, data_content_selection, precondition=None, first=False):
        super().__init__()
       
        self.data_content_selection = data_content_selection
        self.precondition = _time_instance_auto_wrap(precondition)
        self.first = first

    def __str__(self):
        first_text = "first" if self.first else "last"
        precondition_text = f" when {self.precondition}" if self.precondition is not None else ""
        return f"the {first_text} value of {self.data_content_selection}{precondition_text}"

class TimeMeasure(_MeasureDefinition):
    """Measure to calc the time elapsed between A condition and B condition.
    
    Args:
    
        - from_condition: Condition where you want to start counting, defined as A
        - to_condition: End condition, defined as B
        - time_measure_type: Linear in case just want a simple A-B calc, Cyclic in case you want to count 
            all pairs A-B appearances
        - single_instance_agg_function: Only in Cyclic mode. Type of operation applied to our dataset, 
            can be MAX, MIN, SUM, or AVG. 
        - first_to: Only in Linear mode. True to take first appareance of B, false to take last appareance of B
        - precondition: Filter to previusly apply to our dataset. Note that, in this case, this precondition
                        is not a time instant condition, but a data condition (i.e., it is directly checked with
                        the values of the attributes of each event)
        - business_duration: Business days and hours specification
        - time_unit: The time unit used to return the value.
    """

    def __init__(self, from_condition, to_condition, 
                    time_measure_type = 'LINEAR', single_instance_agg_function = 'SUM', 
                         first_to = False, precondition = None, business_duration = None, time_unit = None):
        super().__init__()
  
        self.from_condition = _time_instance_auto_wrap(from_condition)
        self.to_condition = _time_instance_auto_wrap(to_condition)
        self.time_measure_type = time_measure_type
        self.single_instance_agg_function = single_instance_agg_function
        self.precondition = precondition
        self.first_to = first_to
        self.business_duration = business_duration
        self.time_unit = time_unit
        
    def __repr__(self):
        return f"TimeMeasure ( from={self.from_condition}, to={self.to_condition} )"

    def __str__(self):
        precon_text = f" if {self.precondition}" if self.precondition is not None else ""
        first_to_text = "first" if self.first_to else "last"

        if self.time_measure_type.upper() == 'LINEAR':
            text = f"the duration between the first time instant when {self.from_condition} and the {first_to_text} time instant when {self.to_condition}{precon_text}"
        else:
            agg_text = _name_of_aggregation(self.single_instance_agg_function)
            text = f"the {agg_text} duration between the pairs of time instants when {self.from_condition} and when {self.to_condition}{precon_text}"
        return text


class AggregatedMeasure(_MeasureDefinition):
    """Measure to group a dataset by certain time interval.
    
    Args:
    
            - base_measure: First measure applied to the base dataframe.
            - single_instance_agg_function: Type of operation applied to our dataset, 
                can be MAX, MIN, SUM, AVG or GROUPBY.
            - grouper: Array of measures for grouping.
            - filter_to_apply: Measure applied to our base_measure to filter specific values. The measure
                               must resolve to either true or false.
    """
    def __init__(self, base_measure, single_instance_agg_function, grouper=None, filter_to_apply=None):
        super().__init__()
  
        self.base_measure = base_measure
        self.filter_to_apply = filter_to_apply
        self.single_instance_agg_function = single_instance_agg_function
        if grouper is None:
            self.grouper = []
        elif isinstance(grouper, _MeasureDefinition):
            self.grouper = [grouper]
        else:
            self.grouper = grouper

    def __str__(self):
        result =  f"the {_name_of_aggregation(self.single_instance_agg_function)} of {self.base_measure}"
        if len(self.grouper) > 0:
            grouper_text = ", ".join([f"{g}" for g in self.grouper])
            result = f"{result} grouped by {grouper_text}"
        if self.filter_to_apply is not None:
            result = f"{result} filtered by {self.filter_to_apply}"

        return result
        
 
class DerivedMeasure(_MeasureDefinition):
    """Measure that applies an operation to a group of measures.
    
    Args:
    
            - function_expression: Aritmathical or logical expresion as String
            - measure_map: Map of {name_of_measure: Measure}
    """
    
    def __init__(self, function_expression, measure_map):
        super().__init__()
      
        self.function_expression = function_expression
        self.measure_map = measure_map

    def __str__(self):
        variables = ", ".join([f"{m} is {self.measure_map[m]}" for m in self.measure_map])

        return f"the function {self.function_expression} where {variables}"

class BusinessDuration():
    
    def __init__(self, business_start, business_end, weekend_list=[5,6], holiday_list=None, unit_hour='min'):
        super().__init__()
        
        self.business_start = business_start
        self.business_end = business_end
        self.weekend_list = weekend_list
        self.holiday_list = holiday_list
        self.unit_hour = unit_hour

    def conversion(self):
        if(self.unit_hour == 'day'):
            time_delta_type = lambda x: (datetime.timedelta(days = x))
        elif(self.unit_hour == 'hour'):
            time_delta_type = lambda x: (datetime.timedelta(hours = x))
        elif(self.unit_hour == 'min'):
            time_delta_type = lambda x: (datetime.timedelta(minutes = x))
        elif(self.unit_hour == 'sec'):
            time_delta_type = lambda x: (datetime.timedelta(seconds = x))

        return time_delta_type

class RollingWindow():
    """Window applied to the data to group with a moving window.
    
    Args:
        - window: Size of the moving window, could be a number or a time expression ('7D'). If it is a number, then apply_to_cases must be true.
        - apply_to_cases: Returns the moving window for each case instead of based on the window frequency.
        - min_period: Minimum number of observations in window required to have a value 
        - closed: Make the interval closed on the ‘right’, ‘left’, ‘both’ or ‘neither’ endpoints.
    """
    def __init__(self, window, apply_to_cases=False, min_period=None, closed=None):
        if (isinstance(window, int) or (isinstance(window,float) and window.is_integer())) and not apply_to_cases:
            raise ValueError('If window is an integer, apply_to_cases must be True')

        self.window = window
        self.apply_to_cases = apply_to_cases
        self.min_period = min_period
        self.closed = closed



 
        



from pandas._libs.tslibs import NaT
from ppinot4py.model import *
from ppinot4py.computers import *
from datetime import time
import holidays as pyholidays
import datetime
import pandas as pd
import pytest
import numpy as np
import math


@pytest.fixture
def log_for_time():
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    IdCase3 = '3-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2011, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2013, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2014, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2015, 6, 6, 16, 44, 7)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase2, IdCase2, IdCase3, IdCase3],
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)

    return dataframeLinear


@pytest.fixture
def log_config():

    return LogConfiguration()


def test_aggregated_compute_time_grouped(log_for_time, log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA,
        single_instance_agg_function='SUM')

    timeResult1 = datetime.timedelta(days=365, minutes=46, seconds=6)

    var = measure_computer(aggregatedMeasure, log_for_time,
                           log_config, time_grouper=pd.Grouper(freq='1Y'))

    assert var.size == 5
    assert var.iloc[0] == timeResult1
    assert pd.isnull(var.iloc[1])
    assert var.iloc[2] == datetime.timedelta(days=365)




def test_aggregated_compute_time_no_group(log_for_time, log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA,
        single_instance_agg_function='SUM')

    var = measure_computer(aggregatedMeasure, log_for_time, log_config)

    assert var == datetime.timedelta(days=1126, minutes=46, seconds=6)

def test_aggregated_compute_time_no_group_min(log_for_time, log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA,
        single_instance_agg_function='MIN')

    var = measure_computer(aggregatedMeasure, log_for_time, log_config)

    assert var == datetime.timedelta(days=365)    


def test_derived_compute_time_condition(log_for_time, log_config):
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    derivedMeasure = DerivedMeasure("time < days366",
                                    {"time": timeMeasureLinearA,
                                     "days366": pd.Timedelta(days=366)})

    var = measure_computer(derivedMeasure, log_for_time, log_config)

    assert var.size == 3
    assert var.iloc[0]
    assert var.iloc[1]
    assert not var.iloc[2]


def test_derived_aggregated_time(log_for_time, log_config):
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA,
        single_instance_agg_function='SUM')

    derivedMeasure = DerivedMeasure("time < days366",
                                    {"time": aggregatedMeasure,
                                     "days366": pd.Timedelta(days=366)})

    var = measure_computer(derivedMeasure, log_for_time,
                           log_config, time_grouper=pd.Grouper(freq='1Y'))

    assert var.size == 5
    assert np.all(var == [True, False, True, False, False])


def test_count_compute(log_config):

    countMeasure = CountMeasure('`lifecycle:transition` == "In Progress"')

    IdCase1 = '1-364285768'

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(countMeasure, dataframeLinear, log_config).iloc[0]

    assert var == 2


def test_count_compute_not_appear(log_config):

    countMeasure = CountMeasure('`lifecycle:transition` == "Closed"')

    IdCase1 = '1-364285768'

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(countMeasure, dataframeLinear, log_config).iloc[0]

    assert var == 0


def test_count_compute_instances(log_config):

    countMeasure = CountMeasure('`lifecycle:transition` == "In Progress"')

    IdCase1 = '1-364285768'
    idCase2 = '2-364285768'

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, idCase2, idCase2, idCase2],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'In Progress', 'In Progress', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(countMeasure, dataframeLinear, log_config)

    assert var.iloc[0] == 2
    assert var.iloc[1] == 1


def test_data_computer(log_config):

    dataMeasure = DataMeasure(
        data_content_selection="lifecycle:transition",
        first=False)

    IdCase1 = '1-364285768'

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1],
            'concept:name': ['Queued', 'Queued', 'Not queued'],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'Completed']}

    dataframeLinear = pd.DataFrame(data)
    result = measure_computer(dataMeasure, dataframeLinear, log_config)

    assert result.iloc[0] == 'Completed'


def test_data_computer_precondition(log_config):

    dataMeasure = DataMeasure(
        precondition="`concept:name` == 'Queued'",
        data_content_selection="lifecycle:transition",
        first=False)

    IdCase1 = '1-364285768'

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1],
            'concept:name': ['Queued', 'Queued', 'Not queued'],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    result = measure_computer(dataMeasure, dataframeLinear, log_config)

    assert result.iloc[0] == 'In Progress'


def test_time_linear_instances(log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)

    IdCase1 = '1-364285768'

    timeResult = datetime.timedelta(minutes=46, seconds=6)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureLinearA, dataframeLinear, log_config).iloc[0]
    assert var == timeResult


def test_time_linear_instances_with_several_from(log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=False)

    IdCase1 = '1-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(minutes=46, seconds=6)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'In Progress', 'In Progress', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureLinearA, dataframeLinear, log_config).iloc[0]
    assert var == timeResult


def test_time_linear_instances_WithSeveralToAndFirstTo(log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    IdCase1 = '1-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(minutes=46, seconds=6)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'In Progress', 'In Progress', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureLinearA, dataframeLinear, log_config).iloc[0]
    assert var == timeResult


def test_time_linear_instances_WithSeveralToAndNotFirstTo(log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=False)

    IdCase1 = '1-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days=797, hours=23, minutes=44, seconds=25)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureLinearA, dataframeLinear, log_config).iloc[0]
    assert var == timeResult


def test_time_linear_instances_WithSeveralFromAndToAndFirstTo2(log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=False)

    IdCase1 = '1-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days=797, hours=23, minutes=44, seconds=25)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureLinearA, dataframeLinear, log_config).iloc[0]
    assert var == timeResult


def test_time_cyclic_SumInstances(log_config):

    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        single_instance_agg_function='SUM',
        time_measure_type='CYCLIC')

    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days=732, minutes=46, seconds=6)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureCyclic, dataframeLinear, log_config).iloc[0]

    assert var == timeResult


def test_time_cyclic_MaxInstances(log_config):

    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        single_instance_agg_function='MAX',
        time_measure_type='CYCLIC')

    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days=731)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureCyclic, dataframeLinear, log_config).iloc[0]

    assert var == timeResult


def test_time_cyclic_AvgInstances(log_config):

    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        single_instance_agg_function='AVG',
        time_measure_type='CYCLIC')

    IdCase1 = '1-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(
        days=191, hours=12, minutes=11, seconds=31.5)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureCyclic, dataframeLinear, log_config).iloc[0]

    assert var == timeResult


def test_derived_instances(log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)

    timeMeasureLinearB = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Closed"',
        first_to=True)

    timeMeasureLinearC = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Pending"',
        first_to=True)

    measure_dictionary = {'ProgressTime': timeMeasureLinearA,
                          'ClosedTime': timeMeasureLinearB, 'PendingTime': timeMeasureLinearC}
    derived_measure = DerivedMeasure(
        function_expression='(ProgressTime + ClosedTime) + PendingTime',
        measure_map=measure_dictionary)

    IdCase1 = '1-364285768'

    time_result = datetime.timedelta(days=3360, minutes=14, seconds=56)

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2011, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2013, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2014, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2015, 6, 6, 16, 44, 7)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Closed', 'In Progress', 'Pending']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(derived_measure, dataframeLinear, log_config)

    assert var[0] == time_result


def test_time_linear_instances_businessDuration_from_7_to_17(log_config):

    business = BusinessDuration(
        business_start=time(7, 0, 0),
        business_end=time(17, 0, 0),
        weekend_list=[5, 6],
        holiday_list=pyholidays.ES(prov='AN'),
        unit_hour='sec'
    )

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True,
        business_duration=business)

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)

    IdCase1 = '1-364285768'

    timeResult = 18

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureLinearA, dataframeLinear, log_config).iloc[0]
    assert var == timeResult


def test_time_linear_instances_businessDuration_from_7_to_17_severalFrom(log_config):

    business = BusinessDuration(
        business_start=time(7, 0, 0),
        business_end=time(17, 0, 0),
        weekend_list=[5, 6],
        holiday_list=pyholidays.ES(prov='AN'),
        unit_hour='sec'
    )

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=False,
        business_duration=business)

    IdCase1 = '1-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = 18

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'In Progress', 'In Progress', 'In Progress']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureLinearA, dataframeLinear, log_config).iloc[0]
    assert var == timeResult


def test_time_cyclic_SumInstances_businessDuration(log_config):

    business = BusinessDuration(
        business_start=time(7, 0, 0),
        business_end=time(17, 0, 0),
        weekend_list=[5, 6],
        holiday_list=pyholidays.ES(prov='AN'),
        unit_hour='hour'
    )

    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        single_instance_agg_function='SUM',
        time_measure_type='CYCLIC',
        business_duration=business)

    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days=209, hours=14, seconds=18)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureCyclic, dataframeLinear, log_config).iloc[0]

    assert var == timeResult


def test_time_cyclic_MaxInstances_businessDuration(log_config):

    business = BusinessDuration(
        business_start=time(7, 0, 0),
        business_end=time(17, 0, 0),
        weekend_list=[5, 6],
        holiday_list=pyholidays.ES(prov='AN'),
        unit_hour='hour'
    )

    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        single_instance_agg_function='MAX',
        time_measure_type='CYCLIC',
        business_duration=business)

    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days=209, hours=4, minutes=15, seconds=53)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureCyclic, dataframeLinear, log_config).iloc[0]

    assert var == timeResult


def test_time_cyclic_AVGInstances_businessDuration(log_config):

    business = BusinessDuration(
        business_start=time(7, 0, 0),
        business_end=time(17, 0, 0),
        weekend_list=[5, 6],
        holiday_list=pyholidays.ES(prov='AN'),
        unit_hour='hour'
    )

    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        single_instance_agg_function='AVG',
        time_measure_type='CYCLIC',
        business_duration=business)

    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'

    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days=69, hours=20, minutes=40, seconds=6)

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1],
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment']}

    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(
        timeMeasureCyclic, dataframeLinear, log_config).iloc[0]

    assert var == timeResult


def test_data_computer_precondition_predefined_log_values(log_config):

    precondition = TimeInstantCondition(
        "'Awaiting Assignment'", AppliesTo.ACTIVITY, "'Queued'")

    dataMeasure = DataMeasure(
        data_content_selection="lifecycle:transition",
        precondition=precondition,
        first=False)

    IdCase1 = '1-364285768'
    IdCase2 = '1-364285769'

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase2, IdCase1],
            'concept:name': ['Queued', 'Queued', 'Not queued', 'Queued', 'Not queued'],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment', 'Completed', 'Awaiting Assignment', 'Completed']}

    dataframeLinear = pd.DataFrame(data)

    result = measure_computer(dataMeasure, dataframeLinear, log_config)

    assert result.iloc[0] == 'Awaiting Assignment'


def test_data_computer_precondition_non_predefined_log_values(log_config):

    precondition = TimeInstantCondition(
        "'Awaiting Assignment'", AppliesTo.ACTIVITY, "'Queued'")

    log = LogConfiguration(
        id_case='case:concept:name',
        time_column='time:timestamp',
        transition_column='lifecycle:transition:transition',
        activity_column='concept:name:non:predefined')

    dataMeasure = DataMeasure(
        data_content_selection="lifecycle:transition:transition",
        precondition=precondition,
        first=False)

    IdCase1 = '1-364285768'
    IdCase2 = '1-364285769'

    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase2, IdCase1],
            'concept:name:non:predefined': ['Queued', 'Queued', 'Not queued', 'Queued', 'Not queued'],
            'lifecycle:transition:transition': ['In Progress', 'Awaiting Assignment', 'Completed', 'Awaiting Assignment', 'Completed']}

    dataframeLinear = pd.DataFrame(data)

    result = measure_computer(dataMeasure, dataframeLinear, log)

    assert result.iloc[0] == 'Awaiting Assignment'

def test_aggregated_compute_time_rolling_base_date_window_800_days_sum_to_cases(log_for_time, log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA, 
        single_instance_agg_function='SUM')

    rolling = RollingWindow(
      window = "800D",
      apply_to_cases=True
    )
    
    var = measure_computer(aggregatedMeasure, log_for_time, log_config, time_grouper=rolling)
    
    assert var.size == 3
    assert var.iloc[0] == datetime.timedelta(days=365, minutes=46, seconds=6)
    assert var.iloc[1] == datetime.timedelta(days=730, minutes=46, seconds=6)
    assert var.iloc[2] == datetime.timedelta(days=761)

def test_aggregated_compute_time_rolling_base_date_window_800_days_sum(log_for_time, log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA, 
        single_instance_agg_function='SUM')

    rolling = RollingWindow(
      window = "800D"
    )
    
    var = measure_computer(aggregatedMeasure, log_for_time, log_config, time_grouper=rolling)

    counts = var.value_counts()
    
    assert var.size == 1529
    assert counts.index[0] == datetime.timedelta(days=365, minutes=46, seconds=6)
    assert counts.index[1] == datetime.timedelta(days=365)
    assert counts.index[2] == datetime.timedelta(days=730, minutes=46, seconds=6)
    assert counts.index[3] == datetime.timedelta(days=761)

def test_aggregated_compute_time_rolling_base_date_window_800_days_avg(log_for_time, log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA, 
        single_instance_agg_function='AVG')

    rolling = RollingWindow(
      window = "800D"
    )
    
    var = measure_computer(aggregatedMeasure, log_for_time, log_config, time_grouper=rolling)

    counts = var.value_counts()
    
    assert var.size == 1529
    assert counts.index[0] == datetime.timedelta(days=365, minutes=46, seconds=6)
    assert counts.index[1] == datetime.timedelta(days=365)
    assert counts.index[2] == datetime.timedelta(days=365, minutes=23, seconds=3)
    assert counts.index[3] == datetime.timedelta(days=380, hours=12)

def test_aggregated_compute_time_rolling_base_numeric_window_3_center_sum(log_for_time, log_config):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA, 
        single_instance_agg_function='SUM')

    rolling = RollingWindow(
      window = 2,
      apply_to_cases=True
    )
    
    var = measure_computer(aggregatedMeasure, log_for_time, log_config, time_grouper=rolling)
    
    assert var.size == 3
    assert pd.isnull(var.iloc[0])
    assert var.iloc[1] == datetime.timedelta(days=730, minutes=46, seconds=6)
    assert var.iloc[2] == datetime.timedelta(days=761)

def test_time_computer_different_time_units_weeks(log_for_time):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True,
        time_unit=TimeUnit.WEEK)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA,
        single_instance_agg_function='SUM')

    timeResult1 = 52.14743055555556
    timeResult2 = 52.142857142857146

    var = measure_computer(aggregatedMeasure, log_for_time,
                           LogConfiguration(), time_grouper=pd.Grouper(freq='1Y'))

    assert var.size == 5
    assert var.iloc[0] == timeResult1
    assert pd.isnull(var.iloc[1])
    assert var.iloc[2] == timeResult2


def test_time_computer_different_time_units_days(log_for_time):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True,
        time_unit=TimeUnit.DAY)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA,
        single_instance_agg_function='SUM')

    timeResult1 = 365.0320138888889
    timeResult2 = 365.0

    var = measure_computer(aggregatedMeasure, log_for_time,
                           LogConfiguration(), time_grouper=pd.Grouper(freq='1Y'))

    assert var.size == 5
    assert var.iloc[0] == timeResult1
    assert pd.isnull(var.iloc[1])
    assert var.iloc[2] == timeResult2




import datetime
import pandas as pd
import pytest
from ppinot4py.model import RuntimeState, TimeInstantCondition, AppliesTo
from ppinot4py.computers import condition_computer
from business_duration import businessDuration
import holidays as pyholidays
from datetime import time
from itertools import repeat


def test_time_instant_process_start():
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)

    cond = TimeInstantCondition(RuntimeState.START, AppliesTo.PROCESS)
    result = condition_computer(dataframeLinear, "case:concept:name", cond, 'lifecycle:transition', 'concept:name')

    result_expected = [True, False, False, False, False, False, True, False]

    assert (result.values == result_expected).all()
    

def test_time_instant_process_end():
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)

    cond = TimeInstantCondition(RuntimeState.END, AppliesTo.PROCESS)
    result = condition_computer(dataframeLinear, "case:concept:name", cond, 'lifecycle:transition', 'concept:name')

    result_expected = [False, False, False, False, False, False, True, True]

    assert (result.values == result_expected).all()


def test_time_instant_activity_start():

        IdCase1 = '1-364285768'
        IdCase2 = '1-364285769'

        time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
        time1b = datetime.datetime(2010, 3, 31, 17, 32, 48)
        time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
        time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
        time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
        time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)

        data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1], 
                'concept:name': ['Queued', 'Queued', 'Queued', 'Not queued', 'Queued', 'Not queued'],
                'time:timestamp': [time1, time1b, time2, time3, time4, time5],
                'lifecycle:transition': ['Assigned', 'Assigned', 'Awaiting Assignment','Completed', 'Assigned','Completed']}
        
        dataframeLinear = pd.DataFrame(data)

        precondition = TimeInstantCondition(RuntimeState.START, AppliesTo.ACTIVITY, "'Queued'")
        result = condition_computer(dataframeLinear, "case:concept:name", precondition, 'concept:name', 'lifecycle:transition')
        
        result_expected = [True, False, False, False, True, False]

        assert (result.values == result_expected).all()


def test_time_instant_activity_end():

        IdCase1 = '1-364285768'
        IdCase2 = '1-364285769'

        time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
        time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
        time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
        time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
        time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)

        data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase2, IdCase2], 
                'concept:name': ['Queued', 'Queued', 'Not queued', 'Queued', 'Not queued'],
                'time:timestamp': [time1, time2, time3, time4, time5],
                'lifecycle:transition': ['Assigned', 'Awaiting Assignment','Completed', 'Assigned','Completed']}
        
        dataframeLinear = pd.DataFrame(data)

        precondition = TimeInstantCondition(RuntimeState.END, AppliesTo.ACTIVITY, "'Not queued'")
        result = condition_computer(dataframeLinear, "case:concept:name", precondition, 'concept:name', 'lifecycle:transition')
        
        result_expected = [False, False, True, False, True]

        assert (result.values == result_expected).all()


def test_time_instant_activity_custom():

        IdCase1 = '1-364285768'
        IdCase2 = '1-364285769'

        time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
        time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
        time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
        time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
        time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
        time6 = datetime.datetime(2013, 5, 1, 16, 44, 7)

        data = {'case:concept:name': [IdCase1, IdCase1, IdCase1, IdCase2, IdCase2, IdCase2], 
                'concept:name': ['Queued', 'Queued', 'Not queued', 'Queued', 'Queued', 'Not queued'],
                'time:timestamp': [time1, time2, time3, time4, time5, time6],
                'lifecycle:transition': ['Assigned', 'Awaiting Assignment','Completed', 'Assigned', 'Awaiting Assignment', 'Completed']}
        
        dataframeLinear = pd.DataFrame(data)

        precondition = TimeInstantCondition("'Awaiting Assignment'", AppliesTo.ACTIVITY, "'Queued'")
        result = condition_computer(dataframeLinear, "case:concept:name", precondition, 'concept:name', 'lifecycle:transition')
        
        result_expected = [False, True, False, False, True, False]

        assert (result.values == result_expected).all()

def test_time_instant_process_activity_no_lifecycle_END():

        IdCase1 = '1-364285768'
        IdCase2 = '1-364285769'
        IdCase3 = '1-364285770'
        IdCase4 = '1-364285771'
        IdCase5 = '1-364285772'

        time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
        time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
        time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
        time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
        time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)

        data = {'case:concept:name': [IdCase1, IdCase2, IdCase3, IdCase4, IdCase5], 
                'concept:name': ['Queued', 'Queued', 'Not queued', 'Queued', 'Not queued'],
                'time:timestamp': [time1, time2, time3, time4, time5]
        }

        dataframeLinear = pd.DataFrame(data)

        precondition = TimeInstantCondition(RuntimeState.END, AppliesTo.ACTIVITY, "'Queued'")
        result = condition_computer(dataframeLinear, "case:concept:name", precondition, 'concept:name', 'lifecycle:transition')

        result_expected = [True, True, False, True, False]

        assert (result.values == result_expected).all()

def test_time_instant_process_activity_no_lifecycle_START_error():

        IdCase1 = '1-364285768'
        IdCase2 = '1-364285769'
        IdCase3 = '1-364285770'
        IdCase4 = '1-364285771'
        IdCase5 = '1-364285772'

        time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
        time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
        time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
        time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
        time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)

        data = {'case:concept:name': [IdCase1, IdCase2, IdCase3, IdCase4, IdCase5], 
                'concept:name': ['Queued', 'Queued', 'Not queued', 'Queued', 'Not queued'],
                'time:timestamp': [time1, time2, time3, time4, time5]
        }

        dataframeLinear = pd.DataFrame(data)

        precondition = TimeInstantCondition(RuntimeState.START, AppliesTo.ACTIVITY, "'Queued'")

        with pytest.raises(ValueError):
                condition_computer(dataframeLinear, "case:concept:name", precondition, 'concept:name', 'lifecycle:transition')

        



from ppinot4py.model import *
import pytest

def test_count_text():
    c = CountMeasure("ACT == 'open'")
    assert f"{c}" == "the number of times ACT == 'open'"

def test_linear_time_text():
    t = TimeMeasure("ACT == 'open'", "ACT == 'close'")
    assert f"{t}" == "the duration between the first time instant when ACT == 'open' and the last time instant when ACT == 'close'"

def test_data_text():
    d = DataMeasure("AMOUNT")
    assert f"{d}" == "the last value of AMOUNT"

def test_aggregated_text():
    c = CountMeasure("ACT == 'open'")
    a = AggregatedMeasure(c, 'AVG', DataMeasure('PROJECT'))
    assert f"{a}" == "the average of the number of times ACT == 'open' grouped by the last value of PROJECT"

def test_derived_text():
    c = CountMeasure("ACT == 'open'")
    t = TimeMeasure("ACT == 'open'", "ACT == 'close'", first_to=True)
    d = DerivedMeasure("c/t", {"c": c, "t": t})
    assert f"{d}" == "the function c/t where c is the number of times ACT == 'open', t is the duration between the first time instant when ACT == 'open' and the first time instant when ACT == 'close'"

def test_invalid_rolling_window():
    with pytest.raises(ValueError):
        RollingWindow(window=2)
